//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Axis/DiscreteAxis.cpp
//! @brief     Implements class DiscreteAxis.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Base/Axis/DiscreteAxis.h"
#include "Base/Util/Assert.h"
#include "Base/Util/VectorUtil.h"
#include <algorithm>
#include <iomanip>
#include <limits>
#include <stdexcept>

namespace {

std::vector<Bin1D> centers2bins(const std::vector<double>& centers)
{
    std::vector<Bin1D> result;
    for (double c : centers)
        result.emplace_back(Bin1D::At(c));
    return result;
}

} // namespace


DiscreteAxis::DiscreteAxis(const std::string& name, const std::vector<double>& points)
    : IAxis(name, centers2bins(points))
    , m_coordinates(points)
{
    ASSERT(size() == points.size());
    ASSERT(size() == m_coordinates.size());
    sanityCheck();
}

DiscreteAxis::DiscreteAxis(const std::string& name, size_t n, double first, double last)
    : DiscreteAxis(name, VectorUtil::make_grid(n, first, last))
{
}

DiscreteAxis* DiscreteAxis::clone() const
{
    return new DiscreteAxis(axisName(), m_coordinates);
}

void DiscreteAxis::clip(double lower, double upper)
{
    if (lower >= upper)
        throw std::runtime_error("DiscreteAxis::clip() called with invalid lower >= upper");

    using diff_t = std::vector<double>::iterator::difference_type;
    const auto begin = m_coordinates.begin() + static_cast<diff_t>(closestIndex(lower));
    const auto end = m_coordinates.begin() + static_cast<diff_t>(closestIndex(upper)) + 1;

    m_coordinates = std::vector<double>(begin, end);
    sanityCheck();
    m_bins = centers2bins(m_coordinates);
}

void DiscreteAxis::checkIndex(size_t index) const
{
    if (m_coordinates.size() > index)
        return;
    std::string message = "Error in DiscreteAxis::binCenter: passed index ";
    message += std::to_string(index) + " exceeds the size ";
    message += std::to_string(m_coordinates.size()) + " of the axis";
    throw std::runtime_error(message);
}

void DiscreteAxis::sanityCheck() const
{
    if (m_coordinates.size() < 1)
        throw std::runtime_error(
            "Error in DiscreteAxis::DiscreteAxis: the size of passed coordinate array is "
            "less than minimum acceptable value");

    const auto begin = m_coordinates.begin();
    const auto end = m_coordinates.end();

    if (!std::is_sorted(begin, end))
        throw std::runtime_error("Error in DiscreteAxis::DiscreteAxis: passed coordinates are "
                                 "not sorted in ascending order");

    if (std::adjacent_find(begin, end) != end)
        throw std::runtime_error("Error in DiscreteAxis::DiscreteAxis: passed coordinate vector "
                                 "contains repeating values");
}
