//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Axis/IAxis.cpp
//! @brief     Implements interface IAxis.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Base/Axis/IAxis.h"
#include "Base/Util/Algorithms.h"
#include "Base/Util/Assert.h"
#include <iomanip>
#include <optional>

IAxis::~IAxis() = default;

IAxis::IAxis(std::string name, const std::vector<Bin1D>& bins)
    : m_name(name)
    , m_bins(bins)
{
}

size_t IAxis::size() const
{
    return m_bins.size();
}

double IAxis::min() const
{
    return m_bins.front().lowerBound();
}

double IAxis::max() const
{
    return m_bins.back().upperBound();
}

std::pair<double, double> IAxis::bounds() const
{
    return {min(), max()};
}

bool IAxis::rangeComprises(double value) const
{
    return value >= min() && value < max();
}

double IAxis::span() const
{
    return max() - min();
}

double IAxis::center() const
{
    return (max() + min()) / 2;
}

const Bin1D& IAxis::bin(size_t i) const
{
    return m_bins.at(i);
}

double IAxis::binCenter(size_t i) const
{
    return bin(i).center();
}

std::vector<double> IAxis::binCenters() const
{
    std::vector<double> result;
    for (const Bin1D& b : m_bins)
        result.emplace_back(b.center());
    return result;
}

size_t IAxis::closestIndex(double value) const
{
    for (size_t i = 0; i < size() - 1; ++i)
        if (value < (bin(i).upperBound() + bin(i + 1).lowerBound()) / 2)
            return i;
    return size() - 1;
}

bool IAxis::operator==(const IAxis& other) const
{
    return m_name == other.m_name && m_bins == other.m_bins;
}

void IAxis::clip(double /*lower*/, double /*upper*/)
{
    ASSERT(false); // not implemented
}

void IAxis::clip(const std::pair<double, double> bounds)
{
    return clip(bounds.first, bounds.second);
}

std::ostream& operator<<(std::ostream& ostr, const IAxis& ax)
{
    size_t N = ax.size();
    ASSERT(N > 0);

    // Special output for FixedBinAxis:
    bool isFixedBinAxis = true;
    for (size_t i = 0; i < N; ++i) {
        const Bin1D& b = ax.bin(i);
        if (b.lowerBound() != (N - i) * (ax.min() / N) + i * (ax.max() / N)
            || b.upperBound() != (N - i - 1) * (ax.min() / N) + (i + 1) * (ax.max() / N)) {
            isFixedBinAxis = false;
            break;
        }
    }
    if (isFixedBinAxis) {
        ostr << "FixedBinAxis(\"" << ax.axisName() << "\", " << ax.size() << ", "
             << std::setprecision(17) << ax.min() << ", " << ax.max() << ")";
        return ostr;
    }

    // Special output for DiscreteAxis
    bool isDiscreteAxis = true;
    for (const Bin1D& b : ax.bins()) {
        if (b.binSize()) {
            isDiscreteAxis = false;
            break;
        }
    }
    if (isDiscreteAxis) {
        ostr << "DiscreteAxis(\"" << ax.axisName() << "\", "
             << ", [" << std::setprecision(17);
        for (double v : ax.binCenters())
            ostr << v << ",";
        ostr << "])";
        return ostr;
    }

    // Generic output not yet implemented
    ASSERT(false);
}
