#include "Base/Axis/DiscreteAxis.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Device/IO/DataFormatUtil.h"
#include "Tests/GTestWrapper/google_test.h"
#include <iostream>

TEST(PointwiseAxisTest, Construction)
{
    EXPECT_THROW(DiscreteAxis("length", std::vector<double>{1.0, 0.0}), std::runtime_error);
    EXPECT_THROW(DiscreteAxis("length", std::vector<double>{0.0, 1.0, 0.5}), std::runtime_error);
    EXPECT_THROW(DiscreteAxis("length", std::vector<double>{0.0, 1.0, 1.0}), std::runtime_error);
    DiscreteAxis a1("length", std::vector<double>{0.0, 1.0});
    std::vector<double> coordinates{0.0, 1.0};
    DiscreteAxis a2("length", coordinates);
    EXPECT_TRUE(a1 == a2);
}

TEST(PointwiseAxisTest, BasicProperties)
{
    std::vector<double> coordinates{0.0, 1.0, 4.0, 8.0};
    DiscreteAxis a1("length", coordinates);
    EXPECT_EQ(4u, a1.size());
    EXPECT_EQ(0.0, a1.min());
    EXPECT_EQ(8.0, a1.max());
    EXPECT_EQ(0.0, a1.binCenter(0));
    EXPECT_EQ(1.0, a1.binCenter(1));
    EXPECT_EQ(4.0, a1.binCenter(2));
    EXPECT_EQ(8.0, a1.binCenter(3));
    EXPECT_EQ(a1.binCenter(0), a1.binCenter(0));
    EXPECT_EQ(a1.binCenter(1), a1.binCenter(1));
    EXPECT_EQ(a1.binCenter(2), a1.binCenter(2));
    EXPECT_EQ(a1.binCenter(3), a1.binCenter(3));
    EXPECT_THROW(a1.binCenter(4), std::exception);
    EXPECT_THROW(a1.binCenter(4), std::exception);
    EXPECT_TRUE(coordinates == a1.binCenters());
}

TEST(PointwiseAxisTest, FindClosestIndex)
{
    DiscreteAxis v1("name", std::vector<double>{0.0, 1.0, 4.0, 8.0});
    EXPECT_EQ(4u, v1.size());
    EXPECT_EQ(v1.closestIndex(-1.0), 0u);
    EXPECT_EQ(v1.closestIndex(0.0), 0u);
    EXPECT_EQ(v1.closestIndex(0.25), 0u);
    EXPECT_EQ(1u, v1.closestIndex(0.5));
    EXPECT_EQ(1u, v1.closestIndex(0.6));
    EXPECT_EQ(1u, v1.closestIndex(2.49));
    EXPECT_EQ(2u, v1.closestIndex(2.5));
    EXPECT_EQ(3u, v1.closestIndex(8.0));
    EXPECT_EQ(3u, v1.closestIndex(11.0));

    DiscreteAxis v2("name", std::vector<double>{-2.0, -1.0});
    EXPECT_EQ(2u, v2.size());
    EXPECT_EQ(v2.closestIndex(-3.0), 0u);
    EXPECT_EQ(v2.closestIndex(-2.0), 0u);
    EXPECT_EQ(1u, v2.closestIndex(-1.5));
    EXPECT_EQ(1u, v2.closestIndex(-1.0));
    EXPECT_EQ(1u, v2.closestIndex(0.0));
    EXPECT_EQ(1u, v2.closestIndex(1.0));
}

TEST(PointwiseAxisTest, CheckEquality)
{
    DiscreteAxis b1("axis", std::vector<double>{1.0, 2.0, 5.0});
    DiscreteAxis b2("axis", std::vector<double>{1.0, 2.0, 5.0});
    EXPECT_TRUE(b1 == b2);
    DiscreteAxis b3("axissss", std::vector<double>{1.0, 2.0, 5.0});
    DiscreteAxis b4("axis", std::vector<double>{1.0, 2.0, 6.0});
    DiscreteAxis b6("axiss", std::vector<double>{1.5, 2.0, 5.0});
    EXPECT_FALSE(b1 == b3);
    EXPECT_FALSE(b1 == b4);
    EXPECT_FALSE(b1 == b6);
}

TEST(PointwiseAxisTest, CheckClone)
{
    DiscreteAxis a1("axis", std::vector<double>{1.0, 2.0, 5.0});
    std::unique_ptr<DiscreteAxis> clone(a1.clone());
    EXPECT_TRUE(a1 == *clone);
}

TEST(PointwiseAxisTest, IOStream)
{
    DiscreteAxis axis("name", std::vector<double>{1.0, 2.0, 5.0});

    std::ostringstream oss;
    oss << axis;
    std::istringstream iss(oss.str());

    std::unique_ptr<IAxis> result(DataUtils::Format::createAxis(iss));
    EXPECT_TRUE(axis == *result);
}

TEST(PointwiseAxisTest, ClippedAxis)
{
    DiscreteAxis axis("name", std::vector<double>{1.0, 2.0, 2.5, 2.7, 5.0});

    std::unique_ptr<DiscreteAxis> clip1(axis.clone());
    clip1->clip(0.99, 5.1);
    EXPECT_TRUE(*clip1 == axis);

    std::unique_ptr<DiscreteAxis> clip2(axis.clone());
    clip2->clip(1, 5.0);
    EXPECT_TRUE(*clip2 == axis);

    std::unique_ptr<DiscreteAxis> clip3(axis.clone());
    clip3->clip(1.5, 2.5);
    EXPECT_TRUE(*clip3 != axis);
    EXPECT_EQ(clip3->size(), 2u);
    EXPECT_EQ(clip3->binCenter(0), 2.0);
    EXPECT_EQ(clip3->binCenter(1), 2.5);

    // TODO axis restore EXPECT_THROW(axis.clone()->clip(1.4, 1.6), std::runtime_error);
    EXPECT_THROW(axis.clone()->clip(5.0, 1.0), std::runtime_error);
}

TEST(PointwiseAxisTest, FixedBinAxisComparison)
{
    FixedBinAxis fixed_axis("name", 4, 0.0, 4.0);
    DiscreteAxis pointwise_axis("name", std::vector<double>{0.5, 1.5, 2.5, 3.5});

    EXPECT_EQ(fixed_axis.size(), pointwise_axis.size());
    EXPECT_DOUBLE_EQ(fixed_axis.binCenter(0), pointwise_axis.binCenter(0));
    EXPECT_DOUBLE_EQ(fixed_axis.binCenter(1), pointwise_axis.binCenter(1));
    EXPECT_DOUBLE_EQ(fixed_axis.binCenter(2), pointwise_axis.binCenter(2));
    EXPECT_DOUBLE_EQ(fixed_axis.binCenter(3), pointwise_axis.binCenter(3));

    EXPECT_EQ(fixed_axis.closestIndex(0.0), pointwise_axis.closestIndex(0.0));
    EXPECT_EQ(fixed_axis.closestIndex(0.99), pointwise_axis.closestIndex(0.99));
    EXPECT_EQ(fixed_axis.closestIndex(1.0), pointwise_axis.closestIndex(1.0));
    EXPECT_EQ(fixed_axis.closestIndex(2.5), pointwise_axis.closestIndex(2.5));
    EXPECT_EQ(fixed_axis.closestIndex(4.5), pointwise_axis.closestIndex(4.5));

    std::unique_ptr<FixedBinAxis> clipped_fixed(fixed_axis.clone());
    clipped_fixed->clip(0.5, 3.5);
    std::unique_ptr<DiscreteAxis> clipped_pointwise(pointwise_axis.clone());
    clipped_pointwise->clip(0.5, 3.5);
    EXPECT_EQ(clipped_fixed->size(), clipped_pointwise->size());
    EXPECT_DOUBLE_EQ(clipped_fixed->binCenter(0), clipped_pointwise->binCenter(0));
    EXPECT_DOUBLE_EQ(clipped_fixed->binCenter(3), clipped_pointwise->binCenter(3));
}

TEST(PointwiseAxisTest, FixedBinAxisComparisonWithMask)
{
    FixedBinAxis axis("reference", 10, 0.0, 10.0);

    const std::vector<size_t> mask{0u, 2u, 3u, 4u, 7u, 8u, 9u};
    std::vector<double> coordinates;
    for (auto index : mask)
        coordinates.push_back(axis.binCenter(index));
    DiscreteAxis pointwise_axis("pointwise", coordinates);

    // comparing on-axis values
    for (size_t i = 0; i < mask.size(); ++i)
        EXPECT_DOUBLE_EQ(axis.binCenter(mask[i]), pointwise_axis.binCenter(i));
}
